import { Injectable } from '@nestjs/common';
import { InjectEntityManager, InjectRepository } from '@nestjs/typeorm';
import { MaterialsInventoryEntity } from './materials_inventory.entity';
import { EntityManager, Repository } from 'typeorm';
import {
  MaterialsInventoryDto,
  MaterialsInventoryExportDto,
  MaterialsInventoryQueryDto,
  MaterialsInventoryUpdateDto
} from './materials_inventory.dto';
import { Pagination } from '~/helper/paginate/pagination';
import { FastifyReply } from 'fastify';
import { paginate } from '~/helper/paginate';
import * as ExcelJS from 'exceljs';
import dayjs from 'dayjs';
import { MaterialsInOutEntity } from './in_out/materials_in_out.entity';
import { fieldSearch } from '~/shared/database/field-search';
import { groupBy, uniqBy } from 'lodash';
import { MaterialsInOrOutEnum } from '~/constants/enum';
import { ProjectEntity } from '../project/project.entity';
@Injectable()
export class MaterialsInventoryService {
  const;
  constructor(
    @InjectEntityManager() private entityManager: EntityManager,
    @InjectRepository(MaterialsInventoryEntity)
    private materialsInventoryRepository: Repository<MaterialsInventoryEntity>,
    @InjectRepository(MaterialsInOutEntity)
    private materialsInOutRepository: Repository<MaterialsInOutEntity>,
    @InjectRepository(ProjectEntity)
    private projectRepository: Repository<ProjectEntity>
  ) {}

  /**
   * 导出原材料盘点表
   */
  async exportMaterialsInventoryCheck(
    { time, projectId }: MaterialsInventoryExportDto,
    res: FastifyReply
  ): Promise<void> {
    const ROW_HEIGHT = 20;
    const HEADER_FONT_SIZE = 18;
    const workbook = new ExcelJS.Workbook();
    let projects: ProjectEntity[] = [];
    if (projectId) {
      projects = [await this.projectRepository.findOneBy({ id: projectId })];
    }
    // 生成数据
    const sqb = this.materialsInOutRepository
      .createQueryBuilder('mio')
      .leftJoin('mio.project', 'project')
      .leftJoin('mio.product', 'product')
      .leftJoin('product.unit', 'unit')
      .leftJoin('product.company', 'company')
      .addSelect(['project.id','project.name', 'product.name', 'unit.label', 'company.name'])
      .where(fieldSearch({ time }))
      .andWhere('mio.isDelete = 0');

    if (projectId) {
      sqb.andWhere('project.id = :projectId', { projectId });
    }
    const data = await sqb.addOrderBy('mio.time', 'DESC').getMany();
    if (!projectId) {
      projects = uniqBy(
        data.map(item => item.project),
        'id'
      );
    }

    for (const project of projects) {
      const currentProjectData = data.filter(item => item.projectId === project.id);
      const sheet = workbook.addWorksheet(project.name);
      sheet.mergeCells('A1:T1');
      // 设置标题
      sheet.getCell('A1').value = '山东矿机华信智能科技有限公司原材料盘点表';

      // 设置日期
      sheet.mergeCells('A2:B2');
      sheet.getCell('A2').value = `日期：${dayjs(time[0]).format('YYYY年M月')}`;
      // 设置表头
      const headers = [
        '库存编号',
        '公司名称',
        '产品名称',
        '单位',
        '库存数量',
        '单价',
        '金额',
        '',
        '',
        '',
        '',
        '',
        '',
        '',
        '',
        '结存数量',
        '单价',
        '金额',
        '经办人',
        '领料单号',
        '备注'
      ];
      sheet.addRow(headers);
      sheet.addRow([
        '',
        '',
        '',
        '',
        '',
        '',
        '',
        '入库时间',
        '数量',
        '单价',
        '金额',
        '出库时间',
        '数量',
        '单价',
        '金额',
        '',
        '',
        '',
        '',
        '',
        ''
      ]);
      for (let i = 1; i <= 7; i++) {
        sheet.mergeCells(`${String.fromCharCode(64 + i)}3:${String.fromCharCode(64 + i)}4`);
      }
      // 入库
      sheet.mergeCells('H3:K3');
      sheet.getCell('H3').value = '入库';

      // 出库
      sheet.mergeCells('L3:O3');
      sheet.getCell('L3').value = '出库';

      for (let i = 8; i <= 15; i++) {
        sheet.getCell(`${String.fromCharCode(64 + i)}4`).style.fill = {
          type: 'pattern',
          pattern: 'solid',
          fgColor: { argb: 'FFFFC000' }
        };
      }

      for (let i = 16; i <= 21; i++) {
        sheet.mergeCells(`${String.fromCharCode(64 + i)}3:${String.fromCharCode(64 + i)}4`);
      }

      // 固定信息样式设定
      sheet.eachRow((row, index) => {
        row.alignment = { vertical: 'middle', horizontal: 'center' };
        row.font = { bold: true };
        row.height = ROW_HEIGHT;
        if (index >= 3) {
          row.eachCell(cell => {
            cell.border = {
              top: { style: 'thin' },
              left: { style: 'thin' },
              bottom: { style: 'thin' },
              right: { style: 'thin' }
            };
          });
        }
      });

      const groupedData = groupBy<MaterialsInOutEntity>(currentProjectData, 'inventoryNumber');
      let number = 0;
      for (const key in groupedData) {
        // 目前暂定逻辑出库只有一次或者没有出库。不会对一个入库的记录多次出库,故而用find。
        const inRecord = groupedData[key].find(item => item.inOrOut === MaterialsInOrOutEnum.In);
        const outRecord = groupedData[key].find(item => item.inOrOut === MaterialsInOrOutEnum.Out);
        number++;
        sheet.addRow([
          `${inRecord.inventoryNumber}`,
          inRecord.product.company.name,
          inRecord.product.name,
          inRecord.product.unit.label,
          '0',
          '0',
          '0',
          inRecord.time,
          inRecord.quantity,
          parseFloat(`${inRecord.unitPrice || 0}`),
          parseFloat(`${inRecord.amount || 0}`),
          outRecord?.time || '',
          outRecord?.quantity || '',
          parseFloat(`${outRecord.unitPrice || 0}`),
          parseFloat(`${outRecord.amount || 0}`),
          '0',
          '0',
          '0',
          outRecord?.agent,
          outRecord?.issuanceNumber,
          outRecord?.remark
        ]);
      }
      sheet.getCell('A1').font = { size: HEADER_FONT_SIZE };

      // 固定信息样式设定
      sheet.eachRow((row, index) => {
        if (index >= 5) {
          row.alignment = { vertical: 'middle', horizontal: 'center' };
          row.height = ROW_HEIGHT;
          row.eachCell(cell => {
            cell.border = {
              top: { style: 'thin' },
              left: { style: 'thin' },
              bottom: { style: 'thin' },
              right: { style: 'thin' }
            };
          });
        }
      });

      sheet.columns.forEach((column, index: number) => {
        let maxColumnLength = 0;
        const autoWidth = ['B', 'C', 'U'];
        if (String.fromCharCode(65 + index) === 'B') maxColumnLength = 20;
        if (autoWidth.includes(String.fromCharCode(65 + index))) {
          column.eachCell({ includeEmpty: true }, (cell, rowIndex) => {
            if (rowIndex >= 5) {
              const columnLength = `${cell.value || ''}`.length;
              if (columnLength > maxColumnLength) {
                maxColumnLength = columnLength;
              }
            }
          });
          column.width = maxColumnLength < 12 ? 12 : maxColumnLength; // Minimum width of 10
        } else {
          column.width = 12;
        }
      });
    }
    //读取buffer进行传输
    const buffer = await workbook.xlsx.writeBuffer();
    res
      .header('Content-Type', 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet')
      .header(
        'Content-Disposition',
        `attachment; filename="${encodeURIComponent('导出_excel' + new Date().getTime() + '.xls')}"`
      )
      .send(buffer);
  }

  /**
   * 查询所有盘点信息
   */
  async findAll({
    page,
    pageSize
  }: MaterialsInventoryQueryDto): Promise<Pagination<MaterialsInventoryEntity>> {
    const queryBuilder = this.materialsInventoryRepository
      .createQueryBuilder('materialsInventory')
      .where('materialsInventory.isDelete = 0');

    return paginate<MaterialsInventoryEntity>(queryBuilder, {
      page,
      pageSize
    });
  }

  /**
   * 新增
   */
  async create(dto: MaterialsInventoryDto): Promise<void> {
    await this.materialsInventoryRepository.insert(dto);
  }

  /**
   * 更新
   */
  async update(id: number, data: Partial<MaterialsInventoryUpdateDto>): Promise<void> {
    await this.entityManager.transaction(async manager => {
      await manager.update(MaterialsInventoryEntity, id, {
        ...data
      });
    });
  }

  /**
   * 删除
   */
  async delete(id: number): Promise<void> {
    // 合同比较重要，做逻辑删除
    await this.materialsInventoryRepository.update(id, { isDelete: 1 });
  }

  /**
   * 获取单个合同信息
   */
  async info(id: number) {
    const info = await this.materialsInventoryRepository
      .createQueryBuilder('materialsInventory')
      .where({
        id
      })
      .andWhere('materialsInventory.isDelete = 0')
      .getOne();
    return info;
  }
}
